/* --------------------------------------------------------------------------
 * machdep.c:	Copyright (c) Mark P Jones 1991.   All rights reserved.
 *		See goferite.h for details and conditions of use etc...
 *		Gofer version 2.21 November 1991
 *
 *		Last updated 07/11/91 mpj
 *
 * Machine dependent code -- This version for Sun 3/Sun 4 running SUNOS 4.x
 *			     and also for Turbo C version 1.5 running MSDOS
 * ------------------------------------------------------------------------*/

#include "prelude.h"
#include "storage.h"
#include "connect.h"
#include "errors.h"

#if UNIX
#include <sys/ioctl.h>
#include <signal.h>
#endif

#if TURBOC
#include <dos.h>
#include <conio.h>
#include <stdlib.h>
#include <mem.h>

extern unsigned _stklen = 8192;		/* Allocate an 8k stack segment	   */
#endif

/* --------------------------------------------------------------------------
 * Garbage collection notification:
 * ------------------------------------------------------------------------*/

Bool gcMessages = FALSE;		/* TRUE => print GC messages	   */

Void gcStarted() {			/* notify garbage collector start  */
    if (gcMessages) {
	printf("{{Gc");
	fflush(stdout);
    }
}

Void gcScanning() {			/* notify garbage collector scans  */
    if (gcMessages) {
	putchar(':');
	fflush(stdout);
    }
}

Void gcRecovered(recovered)		/* notify garbage collection done  */
Int recovered; {
    if (gcMessages) {
	printf("%d}}",recovered);
	fflush(stdout);
    }
}

Cell *CStackBase;			/* Retain start of C control stack */

Void gcCStack() {			/* Garbage collect elements off    */
    Cell stackTop = NIL;		/* C stack			   */
    Cell *ptr = &stackTop;
#if TURBOC
    if (((long)(ptr) - (long)(CStackBase))&1)
	internal("gcCStack");
#endif
#if UNIX
    if (((long)(ptr) - (long)(CStackBase))&3)
	internal("gcStack");
#endif
    while (ptr<=CStackBase)
	markWithoutMove(*ptr++);
}

/* --------------------------------------------------------------------------
 * Terminal dependent stuff:
 * ------------------------------------------------------------------------*/

#if UNIX
static Bool messedWithTerminal = FALSE;
static struct sgttyb originalTerminal;

Int getTerminalWidth() {		/* determine width of terminal	   */
    static struct winsize terminalSize;
    ioctl(fileno(stdout),TIOCGWINSZ,&terminalSize);
    return (terminalSize.ws_col==0)? 80 : terminalSize.ws_col;
}

Void normalTerminal() {			/* restore terminal initial state  */
    if (messedWithTerminal)
	ioctl(fileno(stdin),TIOCSETP,&originalTerminal);
}

Void noechoTerminal() {
    struct sgttyb term;

    if (!messedWithTerminal) {
	ioctl(fileno(stdin),TIOCGETP,&originalTerminal);
	messedWithTerminal = TRUE;
    }
    ioctl(fileno(stdin),TIOCGETP,&term);
    term.sg_flags |= CBREAK;		/* turn cbreak on, echo off ...	   */
    term.sg_flags &= (~ECHO);
    ioctl(fileno(stdin),TIOCSETP,&term);
}

Int readTerminalChar() {		/* read character from terminal	   */
    return getchar();			/* without echo, assuming that	   */
}					/* noechoTerminal() is active...   */
#endif

#if TURBOC
static Bool terminalEchoReqd = TRUE;

Int getTerminalWidth() {		/* PC screen is fixed 80 chars	   */
    return 80;
}

Void normalTerminal() {			/* restore terminal initial state  */
    terminalEchoReqd = TRUE;
}

Void noechoTerminal() {			/* turn terminal echo on/off	   */
    terminalEchoReqd = FALSE;
}

Int readTerminalChar() {		/* read character from terminal	   */
    if (terminalEchoReqd)
	return getchar();
    else {
	Int c = getch();
	return c=='\r' ? '\n' : c;
    }
}
#endif

/* --------------------------------------------------------------------------
 * Interrupt handling:
 * ------------------------------------------------------------------------*/

static  Bool broken    = FALSE;
static  Bool breakReqd = FALSE;
static  int  ignoreBreak  Args((Void));
#if UNIX
#define ctrlbrk(bh)	signal(SIGINT,bh)
#endif

Bool breakOn(reqd)			/* set break trapping on if reqd,  */
Bool reqd; {				/* or off otherwise, returning old */
    Bool old  = breakReqd;

    breakReqd = reqd;
    if (reqd) {
	if (broken) {			/* repond to break signal received */
	    broken = FALSE;		/* whilst break trap disabled	   */
	    breakHandler();
	}
	ctrlbrk(breakHandler);
    }
    else
	ctrlbrk(ignoreBreak);

    return old;
}

static int ignoreBreak() {             /* record but don't respond to break*/
    ctrlbrk(ignoreBreak);
    broken = TRUE;
    return(1);
}

/* --------------------------------------------------------------------------
 * Chunk `malloc'ation:
 * ------------------------------------------------------------------------*/

Void far *farCalloc(n,s)
Unsigned n,s; {
#if UNIX
    return (Void *)valloc(n * s);
#endif
#if TURBOC
    return farcalloc((unsigned long)n,(unsigned long)s);
#endif
}

/* --------------------------------------------------------------------------
 * Shell escapes:
 * ------------------------------------------------------------------------*/

Int shellEsc(s)				/* run a shell command (or shell)  */
String s; {
#if UNIX
    extern int system Args((char *));
    if (s[0]=='\0')
	s = fromEnv("SHELL","/bin/sh");
#endif
    return system(s);
}

/* --------------------------------------------------------------------------
 * Floating point support:
 * ------------------------------------------------------------------------*/

#ifdef HASFLOATS
static union {
    Float flVal;
    Cell  clVal;
} fudgeCoerce;

Cell mkFloat(fl)
Float fl; {
    fudgeCoerce.flVal = fl;
    return ap(FLOATCELL,fudgeCoerce.clVal);
}

Float floatOf(c)
Cell c; {
    fudgeCoerce.clVal = snd(c);
    return fudgeCoerce.flVal;
}

String floatToString(fl)			/* Make sure that floating   */
Float fl; {					/* point values print out in */
    static char buffer1[32];			/* a form in which they could*/
    static char buffer2[32];			/* also be entered as floats */
    Int i=0, j=0;

    sprintf(buffer1,"%g",fl);
    while (buffer1[i] && strchr("eE.",buffer1[i])==0)
	buffer2[j++] = buffer1[i++];
    if (buffer1[i]!='.') {
        buffer2[j++] = '.';
	buffer2[j++] = '0';
    }
    while (buffer2[j++]=buffer1[i++])
	;
    return buffer2;
}

Float stringToFloat(s)
String s; {
    extern Float atof Args((String));
    return atof(s);
}
#else
Cell mkFloat(fl)
Float fl; {
    internal("mkFloat");
    return NIL;
}

Float floatOf(c)
Cell c; {
    internal("floatOf");
    return 0;
}

String floatToString(fl)
Float fl; {
    internal("floatToString");
    return "";
}

Float stringToFloat(s)
String s; {
    internal("stringToFloat");
    return 0;
}
#endif

/* --------------------------------------------------------------------------
 * Machine dependent control:
 * ------------------------------------------------------------------------*/

static int panic() {			/* exit in a panic, on receipt of  */
    everybody(EXIT);			/* an unexpected signal		   */
    fprintf(stderr,"Unexpected signal\n");
    exit(1);
    return 1;/*NOTREACHED*/
}

Void machdep(what)			/* Handle machine specific	   */
Int what; {				/* initialisation etc..		   */
    switch (what) {
        case MARK    : break;
#if UNIX
        case INSTALL : signal(SIGHUP,panic);
		       signal(SIGQUIT,panic);
		       signal(SIGTERM,panic);
		       signal(SIGSEGV,panic);
		       signal(SIGBUS,panic);
		       break;
#endif
        case RESET   :
	case BREAK   :
	case EXIT    : normalTerminal();
		       break;
    }
}

/*-------------------------------------------------------------------------*/
